"""Model and its methods to access model metadata"""
from abc import ABC, abstractmethod
from ctypes import c_void_p
from typing import List

from ._api import LIBNUX, convert_to_cchar_array
from ._util import list_to_dict
from .errors import InvalidInput
from .tensor import TensorArray, TensorDesc


class Model(ABC):
    """NPU model binary compiled by Renegade compiler"""

    def __init__(self):
        self._named_tensors = {}
        for desc in self.inputs():
            if desc.name:
                self._named_tensors[desc.name] = desc
        for desc in self.outputs():
            if desc.name:
                self._named_tensors[desc.name] = desc

    @abstractmethod
    def _get_model_ref(self) -> c_void_p:
        """
        Returns a raw model pointer

        :return: a raw pointer of a Model
        """

    @property
    def input_num(self) -> int:
        """Number of input tensors of Model"""
        return LIBNUX.nux_input_num(self._get_model_ref())

    @property
    def output_num(self) -> int:
        """Number of output tensors of Model"""
        return LIBNUX.nux_output_num(self._get_model_ref())

    def _input(self, idx) -> TensorDesc:
        return TensorDesc(LIBNUX.nux_input_desc(self._get_model_ref(), idx))

    def input(self, idx) -> TensorDesc:
        """Return tensor description of i-th input tensor of Model"""
        if idx < 0 or self.input_num <= idx:
            raise IndexError('list index out of input tensors')

        return self._input(idx)

    def inputs(self) -> [TensorDesc]:
        """Tensor descriptions of all input tensors of Model"""
        return [self._input(idx) for idx in range(self.input_num)]

    def _output(self, idx) -> TensorDesc:
        return TensorDesc(LIBNUX.nux_output_desc(self._get_model_ref(), idx))

    def output(self, idx) -> TensorDesc:
        """Returns tensor description of i-th output tensor of Model"""
        if idx < 0 or self.output_num <= idx:
            raise IndexError('list index out of output tensors')

        return self._output(idx)

    def outputs(self) -> [TensorDesc]:
        """Tensor descriptions of all output tensors of Model"""
        return [self._output(idx) for idx in range(self.output_num)]

    def _named_tensor_descs(self, names: List[str]) -> List[TensorDesc]:
        try:
            return [self._named_tensors[name] for name in names]
        except KeyError as e:
            raise InvalidInput(f"Tensor '{e.args[0]}' not found in model")

    def allocate_tensors(self, names: List[str]) -> TensorArray:
        """Creates an array of tensors corresponding to tensor names with allocated buffers"""
        tensor_descs = self._named_tensor_descs(names)
        ptrs = convert_to_cchar_array(names)

        return TensorArray(
            LIBNUX.nux_tensor_array_allocate_by_names(self._get_model_ref(), ptrs, len(names)),
            tensor_descs,
            allocated=True,
        )

    def create_tensors(self, names: List[str]) -> TensorArray:
        """Creates an array of tensors corresponding to tensor names without allocated buffers"""
        tensor_descs = self._named_tensor_descs(names)
        ptrs = convert_to_cchar_array(names)

        return TensorArray(
            LIBNUX.nux_tensor_array_create_by_names(self._get_model_ref(), ptrs, len(names)),
            tensor_descs,
            allocated=False,
        )

    def allocate_inputs(self) -> TensorArray:
        """Creates an array of input tensors with allocated buffers"""
        return TensorArray(
            LIBNUX.nux_tensor_array_allocate_inputs(self._get_model_ref()),
            self.inputs(),
            allocated=True,
        )

    def create_inputs(self) -> TensorArray:
        """Creates an array of input tensors without allocated buffers"""
        return TensorArray(
            LIBNUX.nux_tensor_array_create_inputs(self._get_model_ref()),
            self.inputs(),
            allocated=True,
        )

    def allocate_outputs(self) -> TensorArray:
        """Creates an array of output tensors with allocated buffers"""
        return TensorArray(
            LIBNUX.nux_tensor_array_allocate_outputs(self._get_model_ref()),
            self.outputs(),
            allocated=True,
        )

    def create_outputs(self) -> TensorArray:
        """Creates an array of output tensors without allocated buffers"""
        return TensorArray(
            LIBNUX.nux_tensor_array_create_outputs(self._get_model_ref()),
            self.outputs(),
            allocated=True,
        )

    def summary(self) -> str:
        """Returns the summary of this model"""
        return "Inputs:\n{}\nOutputs:\n{}".format(
            list_to_dict(self.inputs()).__repr__(), list_to_dict(self.outputs()).__repr__()
        )

    def print_summary(self):
        """Prints the summary of this model"""
        print(self.summary())



# Environment variable names
ENV_FURIOSA_LOG_DIR = "FURIOSA_LOG_DIR"
ENV_FURIOSA_LOG_LEVEL = "FURIOSA_LOG_LEVEL"
ENV_NPU_DEVNAME = "NPU_DEVNAME"
ENV_XDG_STATE_HOME = "XDG_STATE_HOME"

# Default NPU device
DEFAULT_DEVNAME = "npu0pe0"



import os
from pathlib import Path
import sys

from . import consts


def is_compile_log_enabled() -> bool:
    """
    Return True or False whether the compile log is enabled or not.

    Returns
        True if the compile log is enabled, or False.
    """
    return 'unittest' not in sys.modules.keys()


def xdg_state_home() -> str:
    """
    Return XDG_STATE_HOME which is the base directory of furiosa logs, history, and other states

    Returns:
        Furiosa home directory
    """
    return os.environ.get(consts.ENV_XDG_STATE_HOME, f"{Path.home()}/.local/state/furiosa")


def log_dir() -> str:
    """
    Return FURIOSA_LOG_DIR where the logs are stored.

    Returns:
        The log directory of furiosa sdk
    """
    return os.environ.get(consts.ENV_FURIOSA_LOG_DIR, f"{xdg_state_home()}/logs")


def current_npu_device() -> str:
    """
    Return the current npu device name

    Returns:
        NPU device name
    """
    return os.environ.get(consts.ENV_NPU_DEVNAME, consts.DEFAULT_DEVNAME)



"""Session and its asynchronous API for model inference"""

import ctypes
from ctypes import byref, c_int, c_void_p
from pathlib import Path
import typing
from typing import Dict, List, Optional, Union

import numpy as np
import yaml

from . import envs
from ._api import LIBNUX
from ._api.v1 import convert_to_cchar_array, decref, increase_ref_count, runtime_version
from ._util import dump_info
from .compiler import _model_image, generate_compiler_log_path
from .errors import InvalidInput, UnsupportedTensorType, into_exception, is_err, is_ok
from .model import Model, TensorArray
from .tensor import Tensor, TensorDesc


def _fill_tensor(value: Union[np.ndarray, np.generic], target: Tensor):
    if value.shape != target.shape or value.dtype != target.numpy_dtype:
        raise InvalidInput(
            f"expected {target.shape} ({target.numpy_dtype}) but got {value.shape} ({value.dtype})"
        )

    target.copy_from(value)


def _fill_all_tensors(
    values: Union[np.ndarray, np.generic, TensorArray, List[Union[np.ndarray, np.generic]]],
    targets: TensorArray,
) -> TensorArray:
    """
    Fills `targets` with buffers copied from `values`
    """
    if isinstance(values, (np.ndarray, np.generic)):
        _fill_tensor(values, targets[0])
        return targets

    if isinstance(values, list):
        if len(values) != targets.len:
            raise InvalidInput(
                f"{targets.len} tensors are expected, " f"but {len(values)} tensors are given"
            )

        for value, target in zip(values, targets):
            _fill_tensor(value, target)

        return targets

    if isinstance(values, TensorArray):
        return values

    raise UnsupportedTensorType()


def _create_session_options(
    device: Optional[str] = None,
    worker_num: Optional[int] = None,
    compile_config: Optional[Dict[str, object]] = None,
    compiler_log: Optional[Path] = None,
    input_queue_size: Optional[int] = None,
    output_queue_size: Optional[int] = None,
):
    options: c_void_p = LIBNUX.nux_session_option_create()
    if device:
        LIBNUX.nux_session_option_set_device(options, device.encode())
    if worker_num:
        LIBNUX.nux_session_option_set_worker_num(options, worker_num)
    if compile_config:
        compile_config = yaml.dump(compile_config).encode()
        LIBNUX.nux_session_option_set_compiler_config(options, compile_config)
    if compiler_log:
        compiler_log = str(compiler_log).encode()
        LIBNUX.nux_session_option_set_compiler_log_path(options, compiler_log)
    if input_queue_size:
        LIBNUX.nux_session_option_set_input_queue_size(options, input_queue_size)
    if output_queue_size:
        LIBNUX.nux_session_option_set_output_queue_size(options, output_queue_size)

    return options


class Session(Model):
    """Provides a blocking API to run an inference task with a given model"""

    def __init__(
        self,
        model: Union[bytes, str, Path],
        device: str = None,
        worker_num: int = None,
        compile_config: Dict[str, object] = None,
    ):

        if device is None:
            device = envs.current_npu_device()
        log_path = generate_compiler_log_path()
        options = _create_session_options(device, worker_num, compile_config, log_path)
        model_image = _model_image(model)

        print(f"Using furiosa-compiler {runtime_version()}")
        sess = c_void_p(None)
        err = LIBNUX.nux_session_create(model_image, len(model_image), options, byref(sess))
        if is_err(err):
            dump_info(log_path)
            raise into_exception(err)

        self.ref = sess
        self._as_parameter_ = self.ref

        super().__init__()

    def _get_model_ref(self) -> c_void_p:
        return LIBNUX.nux_session_get_model(self)

    def run(
        self,
        inputs: Union[np.ndarray, np.generic, TensorArray, List[Union[np.ndarray, np.generic]]],
    ) -> TensorArray:
        """
        Runs an inference task with `inputs`

        Args:
            inputs: It can be a single runtime.Tensor, runtime.TensorArray or \
            numpy.ndarray object. Also, you can pass one TensorArray or a \
            list of numpy.ndarray objects.

        Returns:
            Inference output
        """
        _inputs = self.allocate_inputs()
        outputs = self.create_outputs()
        _inputs = _fill_all_tensors(inputs, _inputs)

        err = LIBNUX.nux_session_run(self.ref, _inputs, outputs)

        if is_err(err):
            raise into_exception(err)

        return outputs

    def run_with(
        self, outputs: typing.List[str], inputs: Dict[str, Union[np.ndarray]]
    ) -> TensorArray:
        """
        Runs an inference task with `inputs`

        Args:
            inputs: It can be a single runtime.Tensor, runtime.TensorArray or \
            numpy.ndarray object. Also, you can pass one TensorArray or a \
            list of numpy.ndarray objects.

        Returns:
            Inference output
        """
        # FIXME: outputs=None should be supported in the future
        if outputs is None:
            raise InvalidInput(message="outputs must be given")

        input_names = [name for name in inputs.keys()]
        input_tensors = self.allocate_tensors(input_names)
        output_tensors = self.create_tensors(outputs)
        input_tensors = _fill_all_tensors([value for value in inputs.values()], input_tensors)

        input_names_ptr = convert_to_cchar_array(input_names)
        output_names_ptr = convert_to_cchar_array(outputs)

        err = LIBNUX.nux_session_run_with(
            self.ref,
            input_names_ptr,
            len(input_names),
            output_names_ptr,
            len(outputs),
            input_tensors,
            output_tensors,
        )
        if is_err(err):
            raise into_exception(err)

        return output_tensors

    def close(self):
        """Close the session and release all resources belonging to the session"""
        if self.ref:
            LIBNUX.nux_session_destroy(self.ref)
            self.ref = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()


class CompletionQueue:
    """Receives the completion results asynchronously from AsyncSession"""

    def __init__(self, ref: c_void_p, context_ty: type, output_descs: [TensorDesc]):
        self._as_parameter_ = ref
        self.ref = ref
        self.context_ty = context_ty
        self.output_descs = output_descs
        self.queue_ok = True

    def recv(self, timeout: Optional[int] = None) -> (object, TensorArray):
        """Receives the prediction results which are asynchronously coming from AsyncSession

        If there are already prediction outputs, it will return immediately.
        Otherwise, it will be blocked until the next result are ready.

        If ``timeout`` is set, ``recv()`` will be blocked only until
        the timeout occurs. If timed out, ``recv()`` throws ``QueueWaitTimeout``
        exception.

        If AsyncSession is closed earlier ``recv()`` will throw
        ``SessionTerminated`` exception.

        Args:
            timeout (int): How long to wait before giving up.
            It should be a positive interger in milliseconds.

        Returns:
            A tuple, whose first value is the context value passed \
            when you submit an inference task and the second value \
            is inference output.
        """
        err = c_int(0)
        context_ref = ctypes.py_object(None)
        outputs_ref = c_void_p(None)

        if timeout:
            if timeout < 0:
                raise RuntimeError("the timeout duration must be a positive integer")
            self.queue_ok = LIBNUX.nux_completion_queue_next_timeout(
                self.ref, timeout, byref(context_ref), byref(outputs_ref), byref(err)
            )
        else:
            self.queue_ok = LIBNUX.nux_completion_queue_next(
                self.ref, byref(context_ref), byref(outputs_ref), byref(err)
            )
        context_val = context_ref.value
        decref(context_ref)

        if is_ok(err):
            return context_val, TensorArray(outputs_ref, self.output_descs, allocated=False)

        raise into_exception(err)

    def close(self):
        """Closes this completion queue.

        If it is closed, AsyncSession also will stop working.
        """
        if self.ref:
            LIBNUX.nux_completion_queue_destroy(self.ref)
            self.ref = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        if self.queue_ok:
            return self.recv()

        raise StopIteration()


class AsyncSession(Model):
    """An asynchronous session for a given model allows to submit predictions"""

    def __init__(self, ref: c_void_p):
        self._ref = ref
        self._as_parameter_ = self._ref
        super().__init__()

        self._tensor_buffers = self.allocate_inputs()

    def _get_model_ref(self) -> c_void_p:
        return LIBNUX.nux_async_session_get_model(self)

    def submit(
        self, values: Union[np.ndarray, np.generic, TensorArray], context: object = None
    ) -> None:
        """
        Submit a prediction request

        It immediately returns without blocking the caller, and
        If the prediction is completed, the outputs will be sent to CompletionQueue.

        Args:
            values: Input values
            context: an additional context to identify the prediction request
        """
        _fill_all_tensors(values, self._tensor_buffers)
        # manually increase reference count to keep the context object while running
        increase_ref_count(context)
        err = LIBNUX.nux_async_session_run(self._ref, context, self._tensor_buffers)

        if is_err(err):
            raise into_exception(err)

    def close(self):
        """Closes this session

        After a session is closed, CompletionQueue will return an error
        if CompletionQueue.recv() is called.
        """
        if self._ref:
            LIBNUX.nux_async_session_destroy(self._ref)
            self._ref = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()


def create(
    model: Union[bytes, str, Path],
    device: str = None,
    worker_num: int = None,
    compile_config: Dict[str, object] = None,
) -> Session:
    """Creates a session for a model

    Args:
        model (bytes, str, Path): a byte string containing a model image or \
        a path string of a model image file
        device: NPU device (str) (e.g., npu0pe0, npu0pe0-1)
        worker_num: Number of workers
        compile_config (Dict[str, object]): Compile config

    Returns:
        the session for a given model, allowing to run predictions. \
        Session is a thread safe.
    """
    return Session(model, device, worker_num, compile_config)


def create_async(
    model: Union[bytes, str, Path],
    context_ty: type = None,
    device: str = None,
    worker_num: int = None,
    input_queue_size: int = None,
    output_queue_size: int = None,
    compile_config: Dict[str, object] = None,
) -> (AsyncSession, CompletionQueue):
    """Creates a pair of the asynchronous session and the completion queue for a given model

    Args:
        model (bytes, str, Path): a byte string containing a model image or \
        a path string of a model image file
        context_ty (type): Type for passing context from AsyncSession to CompletionQueue
        device: NPU device (str) (e.g., npu0pe0, npu0pe0-1)
        worker_num: Number of workers
        input_queue_size: The input queue size, and it must be > 0 and < 2^31.
        output_queue_size: The output queue size, and it must be be > 0 and < 2^31.
        compile_config (Dict[str, object]): Compile config

    Returns:
        A pair of the asynchronous session and the completion queue. \
        the asynchronous session for a given model allows to submit predictions. \
        the completion queue allows users to receive the prediction outputs \
        asynchronously.
    """

    try:
        if device is None:
            device = envs.current_npu_device()

        model_image = _model_image(model)
        log_path = generate_compiler_log_path()
        options = _create_session_options(
            device, worker_num, compile_config, log_path, input_queue_size, output_queue_size
        )

        print(f"Using furiosa-compiler {runtime_version()}")
        sess_ref = c_void_p(None)
        queue_ref = c_void_p(None)
        err = LIBNUX.nux_async_session_create(
            model_image, len(model_image), options, byref(sess_ref), byref(queue_ref)
        )
        if is_ok(err):
            sess = AsyncSession(sess_ref)
            return sess, CompletionQueue(queue_ref, context_ty, sess.outputs())
        else:
            dump_info(log_path)
            raise into_exception(err)
    finally:
        pass



#!/usr/bin/env python

import logging
import os
from os.path import dirname

from setuptools import find_namespace_packages, setup

logger = logging.getLogger(__name__)

version = "0.6.0"

my_dir = dirname(__file__)

here = os.path.abspath(os.path.dirname(__file__))


def git_version(version_: str) -> str:
    """
    Return a version to identify the state of the underlying git repo. The version will
    indicate whether the head of the current git-backed working directory is tied to a
    release tag or not : it will indicate the former with a 'release:{version}' prefix
    and the latter with a 'dev0' prefix. Following the prefix will be a sha of the current
    branch head. Finally, a "dirty" suffix is appended to indicate that uncommitted
    changes are present.
    :param str version_: Semver version
    :return: Found Furiosa SDK version in Git repo
    :rtype: str
    """
    try:
        import git  # pylint: disable=import-outside-toplevel

        try:
            repo = git.Repo(os.path.join(*[my_dir, '..', '..', '.git']))
        except git.NoSuchPathError:
            logger.warning('.git directory not found: Cannot compute the git version')
            return ''
        except git.InvalidGitRepositoryError:
            logger.warning('Invalid .git directory not found: Cannot compute the git version')
            return ''
    except ImportError:
        logger.warning('gitpython not found: Cannot compute the git version.')
        return ''
    if repo:
        sha = repo.head.commit.hexsha
        if repo.is_dirty():
            return f'.dev0+{sha}.dirty'
        # commit is clean
        return f'.release:{version_}+{sha}'
    return 'no_git_version'


def write_version(filename: str = os.path.join(*[my_dir, "furiosa/runtime", "git_version"])):
    """
    Write the Semver version + git hash to file, e.g. ".dev0+2f635dc265e78db6708f59f68e8009abb92c1e65".
    :param str filename: Destination file to write
    """
    text = f"{git_version(version)}"
    with open(filename, 'w', encoding='ascii') as file:
        file.write(text)


if __name__ == "__main__":
    setup_kwargs = {}

    write_version()

    setup(
        version=version,
        packages=find_namespace_packages(include=["furiosa.*"]),
        **setup_kwargs,
    )



"""C Native library binding"""

import ctypes
from ctypes import POINTER, c_bool, c_char_p, c_int, c_ulonglong, c_void_p
import logging
import os
from typing import List

from furiosa.common.native import LogLevel, find_native_libs
from furiosa.runtime import consts

LOG = logging.getLogger(__name__)


def _nux_log_level_from_env() -> int:
    level = os.environ.get(consts.ENV_FURIOSA_LOG_LEVEL, 'INFO')
    return LogLevel[level.upper()].value


def convert_to_cchar_array(list: List[str]):
    """
    Convert List[str] to *const *const char of ctypes.

    This function creates an array of POINTER(c_char_p) whose length is
    len(list) + 1, and fill the array of pointers with bytes. The reason
    why we append one more element is to add a null pointer to the end
    of the list. Then, C, C++ side will be able to recognize where is
    the end of the list without length.
    """
    bytes_list = [bytes(str, 'utf-8') for str in list]
    # create an array of ctypes. Please refer to https://docs.python.org/3/library/ctypes.html#arrays.
    ptrs_list = (ctypes.c_char_p * (len(bytes_list) + 1))()
    ptrs_list[:-1] = bytes_list
    return ptrs_list


# Definition of Session Native C API
LIBNUX = find_native_libs("nux")

LIBNUX.nux_session_option_create.argtypes = []
LIBNUX.nux_session_option_create.restype = c_void_p

LIBNUX.nux_session_option_set_device.argtypes = [c_void_p, c_char_p]
LIBNUX.nux_session_option_set_device.restype = None

LIBNUX.nux_session_option_set_compiler_config.argtypes = [c_void_p, c_char_p]
LIBNUX.nux_session_option_set_compiler_config.restype = c_int

LIBNUX.nux_session_option_set_compiler_log_path.argtypes = [c_void_p, c_char_p]
LIBNUX.nux_session_option_set_compiler_log_path.restype = None

LIBNUX.nux_session_option_set_input_queue_size.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_session_option_set_input_queue_size.restype = None

LIBNUX.nux_session_option_set_output_queue_size.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_session_option_set_output_queue_size.restype = None

LIBNUX.nux_session_option_set_worker_num.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_session_option_set_worker_num.restype = None

LIBNUX.nux_session_option_destroy.argtypes = [c_void_p]
LIBNUX.nux_session_option_destroy.restype = None

LIBNUX.nux_input_num.argtypes = [c_void_p]
LIBNUX.nux_input_num.restype = c_int

LIBNUX.nux_output_num.argtypes = [c_void_p]
LIBNUX.nux_output_num.restype = c_int

LIBNUX.nux_input_desc.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_input_desc.restype = c_void_p

LIBNUX.nux_output_desc.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_output_desc.restype = c_void_p

LIBNUX.nux_tensor_name.argtypes = [c_void_p]
LIBNUX.nux_tensor_name.restype = c_void_p

LIBNUX.nux_tensor_dim_num.argtypes = [c_void_p]
LIBNUX.nux_tensor_dim_num.restype = c_ulonglong

LIBNUX.nux_tensor_dim.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_tensor_dim_num.restype = c_ulonglong

LIBNUX.nux_tensor_axis.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_tensor_axis.restype = c_int

LIBNUX.nux_tensor_dtype.argtypes = [c_void_p]
LIBNUX.nux_tensor_dtype.restype = c_int

LIBNUX.nux_session_create.argtypes = [c_void_p, c_ulonglong, c_void_p, POINTER(c_void_p)]
LIBNUX.nux_session_create.restype = c_int

LIBNUX.nux_session_get_model.argtypes = [c_void_p]
LIBNUX.nux_session_get_model.restype = c_void_p

LIBNUX.nux_session_run.argtypes = [c_void_p, c_void_p, c_void_p]
LIBNUX.nux_session_run.restype = c_int

LIBNUX.nux_session_run_with.argtypes = [
    c_void_p,
    POINTER(c_char_p),
    c_ulonglong,
    POINTER(c_char_p),
    c_ulonglong,
    c_void_p,
    c_void_p,
]
LIBNUX.nux_session_run_with.restype = c_int

LIBNUX.nux_session_destroy.argtypes = [c_void_p]
LIBNUX.nux_session_destroy.restype = None

LIBNUX.nux_async_session_create.argtypes = [
    c_void_p,
    c_ulonglong,
    c_void_p,
    POINTER(c_void_p),
    POINTER(c_void_p),
]
LIBNUX.nux_async_session_create.restype = c_int

LIBNUX.nux_async_session_get_model.argtypes = [c_void_p]
LIBNUX.nux_async_session_get_model.restype = c_void_p

LIBNUX.nux_async_session_run.argtypes = [c_void_p, ctypes.py_object, c_void_p]
LIBNUX.nux_async_session_run.restype = c_int

LIBNUX.nux_async_session_destroy.argtypes = [c_void_p]
LIBNUX.nux_async_session_destroy.restype = None

LIBNUX.nux_completion_queue_next.argtypes = [
    c_void_p,
    POINTER(ctypes.py_object),
    POINTER(c_void_p),
    POINTER(c_int),
]
LIBNUX.nux_completion_queue_next.restype = c_bool

LIBNUX.nux_completion_queue_next_timeout.argtypes = [
    c_void_p,
    c_ulonglong,
    POINTER(ctypes.py_object),
    POINTER(c_void_p),
    POINTER(c_int),
]
LIBNUX.nux_completion_queue_next_timeout.restype = c_bool

LIBNUX.nux_tensor_array_create_by_names.argtypes = [c_void_p, POINTER(ctypes.c_char_p), c_ulonglong]
LIBNUX.nux_tensor_array_create_by_names.restype = c_void_p

LIBNUX.nux_tensor_array_allocate_by_names.argtypes = [
    c_void_p,
    POINTER(ctypes.c_char_p),
    c_ulonglong,
]
LIBNUX.nux_tensor_array_allocate_by_names.restype = c_void_p

LIBNUX.nux_tensor_array_create_inputs.argtypes = [c_void_p]
LIBNUX.nux_tensor_array_create_inputs.restype = c_void_p

LIBNUX.nux_tensor_array_allocate_inputs.argtypes = [c_void_p]
LIBNUX.nux_tensor_array_allocate_inputs.restype = c_void_p

LIBNUX.nux_tensor_array_create_outputs.argtypes = [c_void_p]
LIBNUX.nux_tensor_array_create_outputs.restype = c_void_p

LIBNUX.nux_tensor_array_allocate_outputs.argtypes = [c_void_p]
LIBNUX.nux_tensor_array_allocate_outputs.restype = c_void_p

LIBNUX.nux_tensor_buffer_size.argtypes = [c_void_p]

LIBNUX.nux_tensor_array_len.argtypes = [c_void_p]
LIBNUX.nux_tensor_array_len.restype = c_ulonglong

LIBNUX.nux_tensor_array_get.argtypes = [c_void_p, c_ulonglong]
LIBNUX.nux_tensor_array_get.restype = c_void_p

LIBNUX.nux_tensor_array_destroy.argtypes = [c_void_p]
LIBNUX.nux_tensor_array_destroy.restype = None

LIBNUX.tensor_set_buffer.argtypes = [c_void_p, POINTER(ctypes.c_uint8), c_ulonglong]
LIBNUX.tensor_set_buffer.restype = c_int

LIBNUX.tensor_fill_buffer.argtypes = [c_void_p, POINTER(ctypes.c_uint8), c_ulonglong]
LIBNUX.tensor_fill_buffer.restype = c_int

LIBNUX.nux_tensor_len.argtypes = [c_void_p]
LIBNUX.nux_tensor_len.restype = c_ulonglong

LIBNUX.nux_tensor_size.argtypes = [c_void_p]
LIBNUX.nux_tensor_size.restype = c_ulonglong

LIBNUX.tensor_get_buffer.argtypes = [
    c_void_p,
    POINTER(POINTER(ctypes.c_uint8)),
    POINTER(c_ulonglong),
]
LIBNUX.tensor_get_buffer.restype = c_int

LIBNUX.nux_buffer_destroy.argtypes = [POINTER(ctypes.c_uint8), c_ulonglong]
LIBNUX.nux_buffer_destroy.restype = None

LIBNUX.nux_string_destroy.argtypes = [c_void_p]
LIBNUX.nux_string_destroy.restype = None

# To control manually the reference count
increase_ref_count = ctypes.pythonapi.Py_IncRef
increase_ref_count.argtypes = [ctypes.py_object]
increase_ref_count.restype = None

decref = ctypes.pythonapi.Py_DecRef
decref.argtypes = [ctypes.py_object]
decref.restype = None

# Enable Furiosa logger
LIBNUX.enable_logging(_nux_log_level_from_env())

# Register Ctrl-C signal handler to interrupt native side for long running job
LIBNUX.register_signal_handler()


def runtime_version() -> str:
    return (
        f"{LIBNUX.version().decode('utf-8')} "
        f"(rev: {LIBNUX.git_short_hash().decode('utf-8')} built at {LIBNUX.build_timestamp().decode('utf-8')})"
    )
